# -*- coding: utf-8 -*-
"""
Created on Sat Sep 07 21:15 2024

Some basic commands to run on the server.

@author: james
"""

import asyncio


async def _run_cmd(*args, stdin='', code_block=False):
    """Helper function to run the command and get the result."""
    proc = await asyncio.create_subprocess_exec(
        *args,
        stdout=asyncio.subprocess.PIPE,
        stderr=asyncio.subprocess.PIPE,
        stdin=asyncio.subprocess.PIPE,
    )
    stdout, stderr = await proc.communicate(input=stdin.encode())
    if stderr != b'':
        raise RuntimeError(stderr.decode('utf-8'))
    if code_block:
        return "```\n" + stdout.decode('utf-8') + "```"
    return stdout.decode('utf-8')


async def uptime(*args) -> str:
    """Get server uptime"""
    return await _run_cmd('uptime')


async def disk(*args) -> str:
    """Return information about disk usage"""
    return await _run_cmd('df', '-h', '/', code_block=True)


async def proc(*args) -> str:
    """Return the top 5 processes by CPU utilisation"""
    return await _run_cmd('ps', 'aux', '--sort=-pcpu', '|', 'head', '-n 6')


async def ping(*args) -> str:
    """Do a few pings to determine network speed"""
    ip = args[0] if len(args) > 0 else "1.1.1.1"
    return await _run_cmd('ping', '-c 4', ip, code_block=True)


async def users(*args) -> str:
    """List all users logged in"""
    return await _run_cmd('users', code_block=True)


async def logins(*args) -> str:
    """List the last 10 logins (or terminal sessions)"""
    return await _run_cmd('head', stdin=await _run_cmd('last'), code_block=True)


async def ip(*args) -> str:
    """Get the host external ip"""
    return await _run_cmd('curl', '-4sS', 'http://checkip.amazonaws.com')


async def throws(*args) -> str:
    """Always throws an error"""
    return await _run_cmd('nonexistent')


SYSTEM_COMMANDS = {
    "uptime": uptime,
    "disk": disk,
    "proc": proc,
    "ping": ping,
    "logins": logins,
    "ip": ip,
    "throws": throws,
    "users": users,
}

if __name__ == "__main__":
    async def main():
        print(await uptime())
        print(await ip())
        #print(await ping("8.8.8.8"))
    asyncio.run(main())



# -*- coding: utf-8 -*-
"""
Created on Sat Sep 07 20:14 2024

Simple parsers to make the log files look pretty.

@author: james
"""


def _log_level_to_emoji(level: str) -> str:
    """Conver the log level to an emoji"""
    l = level.lower().strip()
    if l == "debug":
        return ":bug: "
    elif l == "info":
        return ":mag: "
    elif l == "warning" or l == "warn":
        return ":warning: "
    elif l == "error":
        return ":x: "
    else:
        return ":grey_question: "


def basic(message: str) -> str:
    """Do nothing"""
    return message


def monty(message: str) -> str:
    """Parse message output from monty"""
    comps = message.split(" ")
    if len(comps) < 4:
        return message  # corrupted format
    if "start" in message.lower():
        return _log_level_to_emoji(comps[2]) + ":large_green_circle: " + " ".join(comps[3:])
    elif "end" in message.lower() or "finish" in message.lower():
        return _log_level_to_emoji(comps[2]) + ":octagonal_sign: " + " ".join(comps[3:])
    else:
        return _log_level_to_emoji(comps[2]) + " ".join(comps[3:])


def ssh(message: str) -> str:
    """Parse message from sshd."""
    comps = message.split(" ")
    date = " ".join(comps[0:3])
    log = " ".join(comps[5:])

    icon = " "
    if "accepted publickey" in message.lower():
        # Confirmation of public key
        icon = " :key: "
    elif "session opened" in message.lower():
        # Creation of a new session
        icon = " :satellite_antenna: "
    elif "error: kex_exchange_identification" in message.lower():
        # Probably malicious port scanning
        icon = " :warning: "
    elif "invalid format" in message.lower():
        # Most likely someone using HTTP scanning
        icon = " :military_helmet: "

    return date + icon + log


PARSERS = {
    "basic": basic,
    "monty": monty,
    "ssh": ssh,
}


if __name__ == "__main__":
    print(monty("[2024-09-10 11:28:44,276] INFO Run finished and took 3 seconds"))
    print(ssh("Sep 13 06:17:24 hostname sshd[281443]: Accepted publickey for user from 1.1.1.1 port 22 ssh2: RSA SHA256:HASH"))
    print(ssh("Sep 13 06:17:24 hostname sshd[281443]: pam_unix(sshd:session): session opened for user user(uid=1001) by (uid=0)"))
    print(ssh("Nov 22 11:08:36 hostname sshd[2700138]: error: kex_exchange_identification: banner line contains invalid characters"))
    print(ssh("Nov 25 06:20:08 hostname sshd[2834581]: banner exchange: Connection from 1.1.1.1 port 10301: invalid format"))



# -*- coding: utf-8 -*-
"""
Created on Sat Sep 07 19:07 2024

Basic slack bot to send notifications to specific slack channels.

Configure it to read log files for updates and broadcast messages.

@author: james
"""
import logging
logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)

import os
import tomllib
from slack_bolt.app.async_app import AsyncApp
from slack_bolt.adapter.socket_mode.async_handler import AsyncSocketModeHandler
import asyncio
import re

# Local imports
from . import watch_logs
from .commands import SYSTEM_COMMANDS

app = AsyncApp(token=os.environ["SLACK_BOT_TOKEN"], process_before_response=True)

__VERSION__ = "0.9"
USERNAME = ""  # populated by who_am_i

## Bot commands

logs_running = False
async def logs(*args):
    """Process the logs command"""
    global logs_running
    if logs_running:
        return ":white_check_mark: Loggers are running."
    else:
        # Install the watchers
        await watch_logs.setup_tasks(app, config)
        logs_running = True
        return "Loggers running"


async def help(*args):
    """Returns all available commands"""
    msg = (":bulb: *Usage* @BOT [COMMAND] (args)\n_System commands_\n"
           + "\n- ".join([f"`{c}`: {SYSTEM_COMMANDS[c].__doc__}" for c in SYSTEM_COMMANDS.keys()])
           + "\n_Bot commands_\n"
           + "\n- ".join([f"`{c}`: {BOT_COMMANDS[c].__doc__}" for c in BOT_COMMANDS.keys()])
           )
    return msg


async def version(*args):
    """Returns the current version"""
    return "Geologs version " + __VERSION__ + ". See https://github.com/gyndlf/geologs"


BOT_COMMANDS = {
    "logs": logs,
    "help": help,
    "version": version,
}


async def who_am_i(app: AsyncApp):
    global USERNAME
    api_result = await app.client.auth_test()
    if "user" in api_result.data:
        USERNAME = api_result["user_id"]
        logger.info(f"Slack username: {api_result['user']} ({USERNAME})")
    else:
        raise RuntimeError("Could not determine bot username")


async def get_if_reacted(channel, timestamp) -> bool:
    """Return if the current bot has reacted to this message."""
    api_response = await app.client.reactions_get(channel=channel, timestamp=timestamp)
    if "reactions" in api_response["message"].keys():
        for emoji in api_response["message"]["reactions"]:
            if USERNAME in emoji["users"]:  # check if we have personally reacted
                return True
    return False


def remove_other_mentions(command: str) -> str:
    """Remove all other bot mentions from the string to call multiple bots."""
    return re.sub(r"<.*?>", "", command).strip()


@app.event("app_mention")
async def handle_mentions(event, client, say):  # async function
    # Check if reacted (and so already has been processed)
    if await get_if_reacted(event["channel"], event["ts"]):
        logger.info(f"Already responded to mention.")
        return

    logger.info(f"Bot mentioned: {event['text']}")
    try:
        cmd_raw = remove_other_mentions(event["text"]).split(" ")
        cmd = cmd_raw[0]
        if len(cmd_raw) > 1:
            cmds = cmd_raw[1:]
        else:
            cmds = []
    except IndexError:
        cmd = ''

    if cmd == '':  # you just pinged me
        api_response = await client.reactions_add(
            channel=event["channel"],
            timestamp=event["ts"],
            name="eyes",
        )
        return

    if (cmd not in SYSTEM_COMMANDS) and (cmd not in BOT_COMMANDS):  # unknown command
        logger.info(f"Unknown command '{cmd}'")
        api_response = await client.reactions_add(
            channel=event["channel"],
            timestamp=event["ts"],
            name="thinking_face",
        )
        await say("Unknown command. " + "Available commands are " + ", ".join(
            [f"`{c}`" for c in SYSTEM_COMMANDS.keys()] + [f"`{c}`" for c in BOT_COMMANDS.keys()]
        ))
        return

    # Command accepted. Run it
    api_response = await client.reactions_add(
        channel=event["channel"],
        timestamp=event["ts"],
        name="thumbsup",
    )
    if cmd in SYSTEM_COMMANDS:  # System command
        try:
            stdout = await SYSTEM_COMMANDS[cmd](*cmds)
        except (FileNotFoundError, RuntimeError) as error:
            # There was an issue completing the command
            api_response = await client.reactions_add(
                channel=event["channel"],
                timestamp=event["ts"],
                name="x",
            )
            logger.error(error)
            await say(":exclamation: Uh oh.```" + str(error) + "```")
            return
        await say(stdout)

    elif cmd in BOT_COMMANDS:  # Bot command
        await say(await BOT_COMMANDS[cmd](*cmds))


async def main_async():
    #setup_tasks(config)
    handler = AsyncSocketModeHandler(app, os.environ["SLACK_APP_TOKEN"])
    await who_am_i(app)  # load the bot username
    await handler.start_async()


def main(config_file: str):
    global config
    logger.info("Loading config file from %s", config_file)
    config = tomllib.load(open(config_file, "rb"))
    asyncio.run(main_async())



# -*- coding: utf-8 -*-
"""
Created on Sat Sep 07 17:48 2024

Basic slack bot to send notifications to specific slack channels.

Configure it to read log files for updates and broadcast messages.

@author: james
"""

import os
from slack_bolt import App
from slack_bolt.adapter.socket_mode import SocketModeHandler


#print(os.environ.get('JAMES'))


app = App(token=os.environ.get("SLACK_BOT_TOKEN"))

# async def ack_check_command(body, ack):
#     text = body.get("text")
#     if text is None or len(text) == 0:
#         await ack(f":x: Usage: /check [service]")
#     else:
#         await ack(f"Accepted! (task: {body['text']})")
#
#
# async def check_health(respond, body):
#     await asyncio.sleep(8)
#     await respond(f"Healthy!! (task: {body['text']})")
#
#
# app.command("/check")(
#     # ack() is still called within 3 seconds
#     ack=ack_check_command,
#     # Lazy function is responsible for processing the event
#     lazy=[check_health]
# )

# Listens to incoming messages that contain "hello"
@app.message("hello")
def message_hello(message, say):
    # say() sends a message to the channel where the event was triggered
    #say(f"Hey there <@{message['user']}>!")
    say(
        blocks=[
            {
                "type": "section",
                "text": {"type": "mrkdwn", "text": f"Hey there <@{message['user']}>!"},
                "accessory": {
                    "type": "button",
                    "text": {"type": "plain_text", "text": "Click Me"},
                    "action_id": "button_click"
                }
            }
        ],
        text=f"Hey there <@{message['user']}>!"
    )


@app.action("button_click")
def action_button_click(body, ack, say):
    # Acknowledge the action
    ack()
    say(f"<@{body['user']['id']}> clicked the button")


@app.event("app_mention")
def handle_app_mention_events(body, logger):
    print(body)
    logger.info(body)
    logger("heeonfdf")


@app.event("message")
def handle_message_events(body, logger):
    print(body)
    logger.info(body)


@app.command("/check")
def check_command(ack, respond, command):
    ack()
    respond("Looking good... maybe")
    print(f"Attempting to get the status of the server {command}")


if __name__ == "__main__":
    #app.start(port=int(os.environ.get("PORT", 3000)))
    SocketModeHandler(app, os.environ["SLACK_APP_TOKEN"]).start()



# -*- coding: utf-8 -*-
"""
Created on Sun Sep 08 10:00 2024

Run the bot

@author: james
"""

import sys


try:
    from . import geologs
except ImportError:
    from geologs import geologs


def main():
    config_file = "config.toml"
    for arg in sys.argv[1:]:
        if arg.startswith("--conf="):
            config_file = arg[len("--conf="):]
    geologs.main(config_file)


if __name__ == "__main__":
    main()




# -*- coding: utf-8 -*-
"""
Created on Sun Sep 08 07:58 2024

Command and methods for subscribing to watch a log file

@author: james
"""

import asyncio
from typing import Callable
import os
import logging
from slack_bolt.app.async_app import AsyncApp


logger = logging.getLogger(__name__)

from .parsers import PARSERS


# Channel topic to set
TOPIC = "Listening to logs! :notes:"


async def watch_file(app: AsyncApp, fname: str, channel_id: str, delay: int, parser: Callable[[str], str]):
    """Start a loop to watch and send a message for new lines in a log file."""
    # Open the file
    file = open(fname, 'r')
    st_results = os.stat(fname)  # move to the end
    st_size = st_results[6]
    file.seek(st_size)

    while True:  # loop... but do it asynchronously
        where = file.tell()
        line = file.readline().strip()
        if not line:
            await asyncio.sleep(delay)
            file.seek(where)
        else:
            # Something new. Send a message
            logger.info(f"Log '{fname}' updated, sending '{line}'")
            await app.client.chat_postMessage(
                channel=channel_id,
                text=parser(line),
            )


async def watch_process(app: AsyncApp, cmd: str, channel_id: str, delay: int, parser: Callable[[str], str]):
    """Start a process and parse the output as a message to a specific channel"""
    # Open the process
    proc = await asyncio.create_subprocess_exec(
        *cmd.split(),
        stdout=asyncio.subprocess.PIPE,
        stderr=asyncio.subprocess.PIPE,
    )

    while True:
        stdout_data = await proc.stdout.readline()
        stdout = stdout_data.decode('utf-8').strip()

        if stdout:
            logger.info(f"Log '{cmd}' updated, sending '{stdout}'")
            await app.client.chat_postMessage(
                channel=channel_id,
                text=parser(stdout),
            )
        if proc.returncode is not None:
            logger.info("Command finished.")
            break


async def check_channel(app: AsyncApp, channel_id: str):
    """Check if a channel exists (or if the bot has access to it), and if not creates one"""
    api_result = await app.client.conversations_list()
    channel_id = channel_id.strip("#")
    for channel in api_result["channels"]:
        if channel["id"] == channel_id or channel["name"] == channel_id.replace("#", ""):
            # Valid channel
            api_result = await app.client.conversations_join(
                channel=channel["id"],
            )
            if api_result["channel"]["topic"]["value"] != TOPIC:
                logger.info(f"Updated channel #{channel['name']} topic")
                api_result = await app.client.conversations_setTopic(
                    channel=channel["id"],
                    topic=TOPIC
                )
            return
    # No channel exists. Make one
    api_result = await app.client.conversations_create(
        name=channel_id
    )
    api_result = await app.client.conversations_join(
        channel=api_result["channel"]["id"],
    )
    api_result = await app.client.conversations_setTopic(
        channel=api_result["channel"]["id"],
        topic="Listening to logs! (Rock music :notes:)"
    )


async def validate_task(app: AsyncApp, task: dict) -> bool:
    """Check that all the required information is given."""
    for r in ["channel", "delay", "parser"]:
        if r not in task.keys():
            raise KeyError(f"Missing required field '{r}'")
    if "logfile" in task.keys() != "logcmd" in task.keys():
        raise KeyError("Either 'logfile' or 'logcmd' must be present but not both.")
    if task["parser"] not in PARSERS.keys():
        raise KeyError(f"Unknown parser {task['parser']}. Valid parsers are {PARSERS.keys()}.")
    if not isinstance(task["delay"], int):
        raise TypeError("Delay must be of type int.")
    if "logfile" in task.keys():  # Check file exists if applicable
        if not os.path.exists(task["logfile"]):
            raise FileNotFoundError(f"Could not find log file '{task['logfile']}'")
    await check_channel(app, task["channel"])
    return True


async def setup_tasks(app: AsyncApp, config: dict):
    """Set up the tasks to watch the appropriate log files."""
    tasks = config.keys()
    for task_name in tasks:
        task = config[task_name]
        await validate_task(app, task)
        # install the task
        if "logfile" in task.keys():  # a file based task
            logger.info(f"Installing file task '{task_name}' to watch '{task['logfile']}'")
            api_result = await app.client.chat_postMessage(
                channel=task["channel"],
                text=f":open_file_folder: :file_cabinet: Subscribed to logs from `{task['logfile']}`"
            )
            asyncio.ensure_future(
                watch_file(
                    app=app,
                    fname=task["logfile"],
                    channel_id=task["channel"],
                    delay=task["delay"],
                    parser=PARSERS[task["parser"]],
                )
            )
        else:
            logger.info(f"Installing process task '{task_name}' to watch '{task['logcmd']}'")
            api_result = await app.client.chat_postMessage(
                channel=task["channel"],
                text=f":open_file_folder: :electric_plug: Subscribed to logs from `{task['logcmd']}`"
            )
            asyncio.ensure_future(
                watch_process(
                    app=app,
                    cmd=task["logcmd"],
                    channel_id=task["channel"],
                    delay=task["delay"],
                    parser=PARSERS[task["parser"]],
                )
            )

